# Unit test _create_train_X_y ForecasterDirect
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.exceptions import MissingValuesWarning
from sklearn.linear_model import LinearRegression
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from skforecast.preprocessing import TimeSeriesDifferentiator
from skforecast.preprocessing import RollingFeatures
from skforecast.direct import ForecasterDirect

# Fixtures
from .fixtures_forecaster_direct import data  # to test results when using differentiation


def test_create_train_X_y_ValueError_when_len_y_is_lower_than_maximum_window_size_plus_steps():
    """
    Test ValueError is raised when length of y is lower than maximum window_size 
    plus number of steps included in the forecaster.
    """
    y = pd.Series(np.arange(5), name='y')

    forecaster = ForecasterDirect(LinearRegression(), steps=3, lags=3)
    err_msg = re.escape(
        "Minimum length of `y` for training this forecaster is "
        "6. Reduce the number of "
        "predicted steps, 3, or the maximum "
        "window_size, 3, if no more data is available.\n"
        "    Length `y`: 5.\n"
        "    Max step : 3.\n"
        "    Max window size: 3.\n"
        "    Lags window size: 3.\n"
        "    Window features window size: None."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_train_X_y(y=y)

    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=3)
    forecaster = ForecasterDirect(
        LinearRegression(),  steps=3, lags=2, window_features=rolling
    )
    err_msg = re.escape(
        "Minimum length of `y` for training this forecaster is "
        "6. Reduce the number of "
        "predicted steps, 3, or the maximum "
        "window_size, 3, if no more data is available.\n"
        "    Length `y`: 5.\n"
        "    Max step : 3.\n"
        "    Max window size: 3.\n"
        "    Lags window size: 2.\n"
        "    Window features window size: 3."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_train_X_y(y=y)


def test_create_train_X_y_TypeError_when_exog_is_categorical_of_no_int():
    """
    Test TypeError is raised when exog is categorical with no int values.
    """
    y = pd.Series(np.arange(6))
    exog = pd.Series(['A', 'B', 'C', 'A', 'B', 'C'], name='exog', dtype='category')
    forecaster = ForecasterDirect(LinearRegression(), lags=2, steps=2)

    err_msg = re.escape(
        "Categorical dtypes in exog must contain only integer values. "
        "See skforecast docs for more info about how to include "
        "categorical features https://skforecast.org/"
        "latest/user_guides/categorical-features.html"
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster._create_train_X_y(y=y, exog=exog)


def test_create_train_X_y_MissingValuesWarning_when_exog_has_missing_values():
    """
    Test _create_train_X_y is issues a MissingValuesWarning when exog has missing values.
    """
    y = pd.Series(np.arange(6))
    exog = pd.Series([1, 2, 3, np.nan, 5, 6], name='exog')
    forecaster = ForecasterDirect(LinearRegression(), lags=2, steps=2)

    warn_msg = re.escape(
        "`exog` has missing values. Most machine learning models do "
        "not allow missing values. Fitting the forecaster may fail."
    )
    with pytest.warns(MissingValuesWarning, match = warn_msg):
        forecaster._create_train_X_y(y=y, exog=exog)


@pytest.mark.parametrize(
    "y                        , exog", 
    [(pd.Series(np.arange(50), name='y'), pd.Series(np.arange(10), name='exog')), 
     (pd.Series(np.arange(10), name='y'), pd.Series(np.arange(50), name='exog')), 
     (pd.Series(np.arange(10), name='y'), pd.DataFrame(np.arange(50).reshape(25, 2), columns=['exog_1', 'exog_2'])),
     (pd.Series(np.arange(10), index=pd.date_range(start='2022-01-01', periods=10, freq='1D'), name='y'), 
      pd.Series(np.arange(50), index=pd.date_range(start='2022-01-01', periods=50, freq='1D'), name='exog'))
])
def test_create_train_X_y_ValueError_when_len_y_is_different_from_len_exog(y, exog):
    """
    Test ValueError is raised when length of y is not equal to length exog.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=3)

    len_exog = len(exog)
    len_y = len(y)
    y_index_no_ws = y.index[forecaster.window_size:]
    len_y_no_ws = len(y_index_no_ws)
    err_msg = re.escape(
        f"Length of `exog` must be equal to the length of `y` (if index is "
        f"fully aligned) or length of `y` - `window_size` (if `exog` "
        f"starts after the first `window_size` values).\n"
        f"    `exog`              : ({exog.index[0]} -- {exog.index[-1]})  (n={len_exog})\n"
        f"    `y`                 : ({y.index[0]} -- {y.index[-1]})  (n={len_y})\n"
        f"    `y` - `window_size` : ({y_index_no_ws[0]} -- {y_index_no_ws[-1]})  (n={len_y_no_ws})"
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_train_X_y(y=y, exog=exog)


def test_create_train_X_y_ValueError_when_y_and_exog_have_different_index_but_same_length():
    """
    Test ValueError is raised when y and exog have different index but same length.
    """
    y = pd.Series(np.arange(10), name='y')
    y.index = pd.date_range(start='2022-01-01', periods=10, freq='1D')
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=3)

    err_msg = re.escape(
        "When `exog` has the same length as `y`, the index of "
        "`exog` must be aligned with the index of `y` "
        "to ensure the correct alignment of values." 
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster.fit(
            y=y,
            exog=pd.Series(
                np.arange(10),
                index=pd.RangeIndex(start=0, stop=10, step=1),
                name="exog",
            ),
        )


def test_create_train_X_y_ValueError_when_y_and_exog_have_different_index_and_length_exog_no_window_size():
    """
    Test ValueError is raised when y and exog have different index and
    length exog no window_size.
    """
    y = pd.Series(np.arange(10), name='y')
    y.index = pd.date_range(start='2022-01-01', periods=10, freq='1D')
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=3)

    err_msg = re.escape(
        "When `exog` doesn't contain the first `window_size` observations, "
        "the index of `exog` must be aligned with the index of `y` minus "
        "the first `window_size` observations to ensure the correct "
        "alignment of values."
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster.fit(
            y=y,
            exog=pd.Series(
                np.arange(3, 10),
                index=pd.RangeIndex(start=3, stop=10, step=1),
                name="exog",
            ),
        )


def test_create_train_X_y_output_when_lags_3_steps_1_and_exog_is_None():
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    lags is 3 and steps is 1.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = None

    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=1)
    results = forecaster._create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[2., 1., 0.],
                             [3., 2., 1.],
                             [4., 3., 2.],
                             [5., 4., 3.],
                             [6., 5., 4.],
                             [7., 6., 5.],
                             [8., 7., 6.]], dtype=float),
            index   = pd.RangeIndex(start=3, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3']
        ),
        {1: pd.Series(
                data  = np.array([3., 4., 5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=3, stop=10, step=1),
                name  = "y_step_1"
            )
        },
        None,
        None,
        ['lag_1', 'lag_2', 'lag_3'],
        None,
        None
    )

    forecaster.exog_in_ is False
    forecaster.X_train_direct_exog_names_out_ is None
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    assert results[5] == expected[5]
    assert results[6] == expected[6]


def test_create_train_X_y_output_when_interspersed_lags_steps_2_and_exog_is_None():
    """
    Test output of _create_train_X_y when regressor is LinearRegression, 
    interspersed_lags and steps is 2.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = None

    forecaster = ForecasterDirect(LinearRegression(), lags=[1, 3], steps=2)
    results = forecaster._create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[2., 0.],
                             [3., 1.],
                             [4., 2.],
                             [5., 3.],
                             [6., 4.],
                             [7., 5.]], dtype=float),
            index   = pd.RangeIndex(start=4, stop=10, step=1),
            columns = ['lag_1', 'lag_3']
        ),
        {1: pd.Series(
                data  = np.array([3., 4., 5., 6., 7., 8.], dtype=float), 
                index = pd.RangeIndex(start=3, stop=9, step=1),
                name  = "y_step_1"
            ),
         2: pd.Series(
                data  = np.array([4., 5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=4, stop=10, step=1),
                name  = "y_step_2"
            )
        },
        None,
        None,
        ['lag_1', 'lag_3'],
        None,
        None
    )

    forecaster.exog_in_ is False
    forecaster.X_train_direct_exog_names_out_ is None
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    assert results[5] == expected[5]
    assert results[6] == expected[6]


@pytest.mark.parametrize("dtype", 
                         [float, int], 
                         ids = lambda dt: f'dtype: {dt}')
def test_create_train_X_y_output_when_y_is_series_10_steps_1_and_exog_is_series_of_float_int(dtype):
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=1 
    and exog is a pandas Series of floats or ints.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.Series(np.arange(100, 110), name='exog', dtype=dtype)

    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=1)
    results = forecaster._create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 105.],
                             [5., 4., 3., 2., 1., 106.],
                             [6., 5., 4., 3., 2., 107.],
                             [7., 6., 5., 4., 3., 108.],
                             [8., 7., 6., 5., 4., 109.]]),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'exog_step_1']
        ).astype({'exog_step_1': float}),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "y_step_1"
            )
        },
        ['exog'],
        ['exog'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog_step_1'],
        {'exog': exog.dtypes},
        {'exog': exog.dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_step_1']
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


@pytest.mark.parametrize("dtype", 
                         [float, int], 
                         ids = lambda dt: f'dtype: {dt}')
def test_create_train_X_y_output_when_y_is_series_10_steps_2_and_exog_is_series_of_float_int(dtype):
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=2 
    and exog is a pandas Series of floats or ints.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.Series(np.arange(100, 110), name='exog', dtype=dtype)
    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=2)
    results = forecaster._create_train_X_y(y=y, exog=exog)
    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 105., 106.],
                             [5., 4., 3., 2., 1., 106., 107.],
                             [6., 5., 4., 3., 2., 107., 108.],
                             [7., 6., 5., 4., 3., 108., 109.]], dtype=float),
            index   = pd.RangeIndex(start=6, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'exog_step_1', 'exog_step_2']
        ).astype({'exog_step_1': float, 'exog_step_2': float}),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=9, step=1),
                name  = "y_step_1"
            ),
         2: pd.Series(
                data  = np.array([6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=6, stop=10, step=1),
                name  = "y_step_2"
            )
        },
        ['exog'],
        ['exog'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog_step_1', 'exog_step_2'],
        {'exog': exog.dtypes},
        {'exog': exog.dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_step_1', 'exog_step_2']
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


@pytest.mark.parametrize("datetime_index", 
                         [True, False], 
                         ids = lambda dt: f'datetime_index: {dt}')
@pytest.mark.parametrize("dtype", 
                         [float, int], 
                         ids = lambda dt: f'dtype: {dt}')
def test_create_train_X_y_output_when_y_is_series_10_steps_1_and_exog_is_series_of_float_int_with_no_window_size(datetime_index, dtype):
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=1 
    and exog is a pandas Series of floats or ints and no initial window_size
    observations.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.Series(
        np.arange(105, 110), index=pd.RangeIndex(start=5, stop=10, step=1), 
        name='exog', dtype=dtype
    )

    expected_index = pd.RangeIndex(start=5, stop=10, step=1)
    if datetime_index:
        y.index = pd.date_range(start='2022-01-01', periods=10, freq='D')
        exog.index = pd.date_range(start='2022-01-06', periods=5, freq='D')
        expected_index = pd.date_range(start='2022-01-06', periods=5, freq='D')

    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=1)
    results = forecaster._create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 105.],
                             [5., 4., 3., 2., 1., 106.],
                             [6., 5., 4., 3., 2., 107.],
                             [7., 6., 5., 4., 3., 108.],
                             [8., 7., 6., 5., 4., 109.]]),
            index   = expected_index,
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'exog_step_1']
        ).astype({'exog_step_1': float}),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8., 9.], dtype=float), 
                index = expected_index,
                name  = "y_step_1"
            )
        },
        ['exog'],
        ['exog'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog_step_1'],
        {'exog': exog.dtypes},
        {'exog': exog.dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_step_1']
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


@pytest.mark.parametrize("datetime_index", 
                         [True, False], 
                         ids = lambda dt: f'datetime_index: {dt}')
@pytest.mark.parametrize("dtype", 
                         [float, int], 
                         ids = lambda dt: f'dtype: {dt}')
def test_create_train_X_y_output_when_y_is_series_10_steps_2_and_exog_is_series_of_float_int_with_no_window_size(datetime_index, dtype):
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=2 
    and exog is a pandas Series of floats or ints and no initial window_size
    observations.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.Series(
        np.arange(105, 110), index=pd.RangeIndex(start=5, stop=10, step=1), 
        name='exog', dtype=dtype
    )

    expected_index_1 = pd.RangeIndex(start=5, stop=9, step=1)
    expected_index_2 = pd.RangeIndex(start=6, stop=10, step=1)
    if datetime_index:
        y.index = pd.date_range(start='2022-01-01', periods=10, freq='D')
        exog.index = pd.date_range(start='2022-01-06', periods=5, freq='D')
        expected_index_1 = pd.date_range(start='2022-01-06', periods=4, freq='D')
        expected_index_2 = pd.date_range(start='2022-01-07', periods=4, freq='D')

    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=2)
    results = forecaster._create_train_X_y(y=y, exog=exog)
    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 105., 106.],
                             [5., 4., 3., 2., 1., 106., 107.],
                             [6., 5., 4., 3., 2., 107., 108.],
                             [7., 6., 5., 4., 3., 108., 109.]], dtype=float),
            index   = expected_index_2,
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'exog_step_1', 'exog_step_2']
        ).astype({'exog_step_1': float, 'exog_step_2': float}),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8.], dtype=float), 
                index = expected_index_1,
                name  = "y_step_1"
            ),
         2: pd.Series(
                data  = np.array([6., 7., 8., 9.], dtype=float), 
                index = expected_index_2,
                name  = "y_step_2"
            )
        },
        ['exog'],
        ['exog'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog_step_1', 'exog_step_2'],
        {'exog': exog.dtypes},
        {'exog': exog.dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_step_1', 'exog_step_2']
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


@pytest.mark.parametrize("dtype", 
                         [float, int], 
                         ids = lambda dt: f'dtype: {dt}')
def test_create_train_X_y_output_when_y_is_series_10_steps_1_and_exog_is_dataframe_of_float_int(dtype):
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=1 
    and exog is a pandas DataFrame of floats or ints.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.DataFrame({'exog_1': np.arange(100, 110, dtype=dtype),
                         'exog_2': np.arange(1000, 1010, dtype=dtype)})
    
    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=1)
    results = forecaster._create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 105., 1005.],
                             [5., 4., 3., 2., 1., 106., 1006.],
                             [6., 5., 4., 3., 2., 107., 1007.],
                             [7., 6., 5., 4., 3., 108., 1008.],
                             [8., 7., 6., 5., 4., 109., 1009.]], dtype=float),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'exog_1_step_1', 'exog_2_step_1']
        ).astype({'exog_1_step_1': float, 'exog_2_step_1': float}),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "y_step_1"
            )
        },
        ['exog_1', 'exog_2'],
        ['exog_1', 'exog_2'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog_1_step_1', 'exog_2_step_1'],
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes},
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_step_1', 'exog_step_2']
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


@pytest.mark.parametrize("dtype", 
                         [float, int], 
                         ids = lambda dt: f'dtype: {dt}')
def test_create_train_X_y_output_when_y_is_series_10_steps_3_and_exog_is_dataframe_of_float_int(dtype):
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=3 
    and exog is a pandas DataFrame of floats or ints.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.DataFrame({'exog_1': np.arange(100, 110, dtype=dtype),
                         'exog_2': np.arange(1000, 1010, dtype=dtype)})
    
    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=3)
    results = forecaster._create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4, 3, 2, 1, 0, 105, 1005, 106, 1006, 107, 1007],
                             [5, 4, 3, 2, 1, 106, 1006, 107, 1007, 108, 1008],
                             [6, 5, 4, 3, 2, 107, 1007, 108, 1008, 109, 1009]],
                            dtype=float),
            index = pd.RangeIndex(start=7, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'exog_1_step_1', 'exog_2_step_1', 
                       'exog_1_step_2', 'exog_2_step_2', 
                       'exog_1_step_3', 'exog_2_step_3']
        ).astype({'exog_1_step_1': float, 'exog_2_step_1': float, 
                  'exog_1_step_2': float, 'exog_2_step_2': float, 
                  'exog_1_step_3': float, 'exog_2_step_3': float}),
        {1: pd.Series(
                data  = np.array([5., 6., 7.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=8, step=1),
                name  = "y_step_1"
            ),
         2: pd.Series(
                data  = np.array([6., 7., 8.], dtype=float), 
                index = pd.RangeIndex(start=6, stop=9, step=1),
                name  = "y_step_2"
            ),
         3: pd.Series(
                data  = np.array([7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=7, stop=10, step=1),
                name  = "y_step_3"
            )
        },
        ['exog_1', 'exog_2'],
        ['exog_1', 'exog_2'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
         'exog_1_step_1', 'exog_2_step_1', 
         'exog_1_step_2', 'exog_2_step_2', 
         'exog_1_step_3', 'exog_2_step_3'],
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes},
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == [
        'exog_1_step_1', 'exog_2_step_1', 
        'exog_1_step_2', 'exog_2_step_2', 
        'exog_1_step_3', 'exog_2_step_3'
    ]
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


@pytest.mark.parametrize("exog_values, dtype", 
                         [([True]    , bool), 
                          (['string'], str)], 
                         ids = lambda dt: f'values, dtype: {dt}')
def test_create_train_X_y_output_when_y_is_series_10_steps_1_and_exog_is_series_of_bool_str(exog_values, dtype):
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=1 
    and exog is a pandas Series of bool or str.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.Series(exog_values * 10, name='exog', dtype=dtype)

    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=1)
    results = forecaster._create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0.],
                             [5., 4., 3., 2., 1.],
                             [6., 5., 4., 3., 2.],
                             [7., 6., 5., 4., 3.],
                             [8., 7., 6., 5., 4.]], dtype=float),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']
        ).assign(exog_step_1=exog_values * 5).astype({'exog_step_1': dtype}),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "y_step_1"
            )
        },
        ['exog'],
        ['exog'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog_step_1'],
        {'exog': exog.dtypes},
        {'exog': exog.dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_step_1']
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


@pytest.mark.parametrize("exog_values, dtype", 
                         [([True]    , bool), 
                          (['string'], str)], 
                         ids = lambda dt: f'values, dtype: {dt}')
def test_create_train_X_y_output_when_y_is_series_10_steps_2_and_exog_is_series_of_bool_str(exog_values, dtype):
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=2
    and exog is a pandas Series of bool or str.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.Series(exog_values * 10, name='exog', dtype=dtype)

    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=2)
    results = forecaster._create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0.],
                             [5., 4., 3., 2., 1.],
                             [6., 5., 4., 3., 2.],
                             [7., 6., 5., 4., 3.]], dtype=float),
            index   = pd.RangeIndex(start=6, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']
        ).assign(exog_step_1=exog_values * 4,
                 exog_step_2=exog_values * 4).astype({'exog_step_1': dtype, 
                                                      'exog_step_2': dtype}),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=9, step=1),
                name  = "y_step_1"
            ),
         2: pd.Series(
                data  = np.array([6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=6, stop=10, step=1),
                name  = "y_step_2"
            )
        },
        ['exog'],
        ['exog'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog_step_1', 'exog_step_2'],
        {'exog': exog.dtypes},
        {'exog': exog.dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_step_1', 'exog_step_2']
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


@pytest.mark.parametrize("v_exog_1   , v_exog_2  , dtype", 
                         [([True]    , [False]   , bool), 
                          (['string'], ['string'], str)], 
                         ids = lambda dt: f'values, dtype: {dt}')
def test_create_train_X_y_output_when_y_is_series_10_steps_1_and_exog_is_dataframe_of_bool_str(v_exog_1, v_exog_2, dtype):
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=1 
    and exog is a pandas dataframe with two columns of bool or str.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.DataFrame({'exog_1': v_exog_1 * 10,
                         'exog_2': v_exog_2 * 10})
    
    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=1)
    results = forecaster._create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0.],
                             [5., 4., 3., 2., 1.],
                             [6., 5., 4., 3., 2.],
                             [7., 6., 5., 4., 3.],
                             [8., 7., 6., 5., 4.]]),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']
        ).assign(exog_1_step_1=v_exog_1 * 5,
                 exog_2_step_1=v_exog_2 * 5).astype({'exog_1_step_1': dtype, 
                                                     'exog_2_step_1': dtype}),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "y_step_1"
            )
        },
        ['exog_1', 'exog_2'],
        ['exog_1', 'exog_2'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog_1_step_1', 'exog_2_step_1'],
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes},
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_1_step_1', 'exog_2_step_1']
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


@pytest.mark.parametrize("v_exog_1   , v_exog_2  , dtype", 
                         [([True]    , [False]   , bool), 
                          (['string'], ['string'], str)], 
                         ids = lambda dt: f'values, dtype: {dt}')
def test_create_train_X_y_output_when_y_is_series_10_steps_3_and_exog_is_dataframe_of_bool_str(v_exog_1, v_exog_2, dtype):
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=3 
    and exog is a pandas dataframe with two columns of bool or str.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.DataFrame({'exog_1': v_exog_1 * 10,
                         'exog_2': v_exog_2 * 10})
    
    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=3)
    results = forecaster._create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0.],
                             [5., 4., 3., 2., 1.],
                             [6., 5., 4., 3., 2.]],
                            dtype=float),
            index = pd.RangeIndex(start=7, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']
        ).assign(exog_1_step_1=v_exog_1 * 3,
                 exog_2_step_1=v_exog_2 * 3,
                 exog_1_step_2=v_exog_1 * 3,
                 exog_2_step_2=v_exog_2 * 3,
                 exog_1_step_3=v_exog_1 * 3,
                 exog_2_step_3=v_exog_2 * 3
        ).astype({'exog_1_step_1': dtype, 
                  'exog_2_step_1': dtype, 
                  'exog_1_step_2': dtype, 
                  'exog_2_step_2': dtype, 
                  'exog_1_step_3': dtype,
                  'exog_2_step_3': dtype}
        ),
        {1: pd.Series(
                data  = np.array([5., 6., 7.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=8, step=1),
                name  = "y_step_1"
            ),
         2: pd.Series(
                data  = np.array([6., 7., 8.], dtype=float), 
                index = pd.RangeIndex(start=6, stop=9, step=1),
                name  = "y_step_2"
            ),
         3: pd.Series(
                data  = np.array([7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=7, stop=10, step=1),
                name  = "y_step_3"
            )
        },
        ['exog_1', 'exog_2'],
        ['exog_1', 'exog_2'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
         'exog_1_step_1', 'exog_2_step_1', 
         'exog_1_step_2', 'exog_2_step_2', 
         'exog_1_step_3', 'exog_2_step_3'],
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes},
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == [
        'exog_1_step_1', 'exog_2_step_1', 
        'exog_1_step_2', 'exog_2_step_2', 
        'exog_1_step_3', 'exog_2_step_3'
    ]
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


def test_create_train_X_y_output_when_y_is_series_10_steps_1_and_exog_is_series_of_category():
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=1 
    and exog is a pandas Series of category.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.Series(range(10), name='exog', dtype='category')

    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=1)
    results = forecaster._create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0.],
                             [5., 4., 3., 2., 1.],
                             [6., 5., 4., 3., 2.],
                             [7., 6., 5., 4., 3.],
                             [8., 7., 6., 5., 4.]]),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']
        ).assign(exog_step_1=pd.Categorical(range(5, 10), categories=range(10))),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "y_step_1"
            )
        },
        ['exog'],
        ['exog'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog_step_1'],
        {'exog': exog.dtypes},
        {'exog': exog.dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_step_1']
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


def test_create_train_X_y_output_when_y_is_series_10_steps_2_and_exog_is_series_of_category():
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=2 
    and exog is a pandas Series of category.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.Series(range(10), name='exog', dtype='category')

    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=2)
    results = forecaster._create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0.],
                             [5., 4., 3., 2., 1.],
                             [6., 5., 4., 3., 2.],
                             [7., 6., 5., 4., 3.]]),
            index   = pd.RangeIndex(start=6, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']
        ).assign(exog_step_1=pd.Categorical(range(5, 9), categories=range(10)),
                 exog_step_2=pd.Categorical(range(6, 10), categories=range(10))),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=9, step=1),
                name  = "y_step_1"
            ),
         2: pd.Series(
                data  = np.array([6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=6, stop=10, step=1),
                name  = "y_step_2"
            )
        },
        ['exog'],
        ['exog'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog_step_1', 'exog_step_2'],
        {'exog': exog.dtypes},
        {'exog': exog.dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_step_1', 'exog_step_2']
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


def test_create_train_X_y_output_when_y_is_series_10_steps_1_and_exog_is_dataframe_of_category():
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=1 
    and exog is a pandas DataFrame of category.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.DataFrame({'exog_1': pd.Categorical(range(10)),
                         'exog_2': pd.Categorical(range(100, 110))})
    
    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=1)
    results = forecaster._create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0.],
                             [5., 4., 3., 2., 1.],
                             [6., 5., 4., 3., 2.],
                             [7., 6., 5., 4., 3.],
                             [8., 7., 6., 5., 4.]]),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']
        ).assign(
            exog_1_step_1=pd.Categorical(range(5, 10), categories=range(10)),
            exog_2_step_1=pd.Categorical(range(105, 110), categories=range(100, 110))
        ),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "y_step_1"
            )
        },
        ['exog_1', 'exog_2'],
        ['exog_1', 'exog_2'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog_1_step_1', 'exog_2_step_1'],
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes},
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_1_step_1', 'exog_2_step_1']
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


def test_create_train_X_y_output_when_y_is_series_10_steps_3_and_exog_is_dataframe_of_category():
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=3 
    and exog is a pandas DataFrame of category.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.DataFrame({'exog_1': pd.Categorical(range(10)),
                         'exog_2': pd.Categorical(range(100, 110))})
    
    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=3)
    results = forecaster._create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0.],
                             [5., 4., 3., 2., 1.],
                             [6., 5., 4., 3., 2.]],
                            dtype=float),
            index = pd.RangeIndex(start=7, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']
        ).assign(
            exog_1_step_1=pd.Categorical(range(5, 8), categories=range(10)),
            exog_2_step_1=pd.Categorical(range(105, 108), categories=range(100, 110)),
            exog_1_step_2=pd.Categorical(range(6, 9), categories=range(10)),
            exog_2_step_2=pd.Categorical(range(106, 109), categories=range(100, 110)),
            exog_1_step_3=pd.Categorical(range(7, 10), categories=range(10)),
            exog_2_step_3=pd.Categorical(range(107, 110), categories=range(100, 110))
        ),
        {1: pd.Series(
                data  = np.array([5., 6., 7.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=8, step=1),
                name  = "y_step_1"
            ),
         2: pd.Series(
                data  = np.array([6., 7., 8.], dtype=float), 
                index = pd.RangeIndex(start=6, stop=9, step=1),
                name  = "y_step_2"
            ),
         3: pd.Series(
                data  = np.array([7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=7, stop=10, step=1),
                name  = "y_step_3"
            )
        },
        ['exog_1', 'exog_2'],
        ['exog_1', 'exog_2'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
         'exog_1_step_1', 'exog_2_step_1', 
         'exog_1_step_2', 'exog_2_step_2', 
         'exog_1_step_3', 'exog_2_step_3'],
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes},
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == [
        'exog_1_step_1', 'exog_2_step_1', 
        'exog_1_step_2', 'exog_2_step_2', 
        'exog_1_step_3', 'exog_2_step_3'
    ]
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


def test_create_train_X_y_output_when_y_is_series_10_and_exog_is_dataframe_of_float_int_category_steps_1():
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=1 
    and exog is a pandas DataFrame of float, int and category.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.DataFrame({'exog_1': pd.Series(np.arange(100, 110), dtype=float),
                         'exog_2': pd.Series(np.arange(1000, 1010), dtype=int),
                         'exog_3': pd.Categorical(range(100, 110))})
    
    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=1)
    results = forecaster._create_train_X_y(y=y, exog=exog)        
    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 105., 1005.],
                             [5., 4., 3., 2., 1., 106., 1006.],
                             [6., 5., 4., 3., 2., 107., 1007.],
                             [7., 6., 5., 4., 3., 108., 1008.],
                             [8., 7., 6., 5., 4., 109., 1009.]]),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'exog_1_step_1', 'exog_2_step_1']
        ).astype({'exog_1_step_1': float, 
                  'exog_2_step_1': int}
        ).assign(exog_3_step_1=pd.Categorical(range(105, 110), categories=range(100, 110))
        ),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "y_step_1"
            )
        },
        ['exog_1', 'exog_2', 'exog_3'],
        ['exog_1', 'exog_2', 'exog_3'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
         'exog_1_step_1', 'exog_2_step_1', 'exog_3_step_1'],
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes, 'exog_3': exog['exog_3'].dtypes},
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes, 'exog_3': exog['exog_3'].dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_1_step_1', 'exog_2_step_1', 'exog_3_step_1']
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


def test_create_train_X_y_output_when_y_is_series_10_and_exog_is_dataframe_of_float_int_category_steps_3():
    """
    Test the output of _create_train_X_y when y=pd.Series(np.arange(10)), steps=3 
    and exog is a pandas DataFrame of float, int and category.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.DataFrame({'exog_1': pd.Series(np.arange(100, 110), dtype=float),
                         'exog_2': pd.Series(np.arange(1000, 1010), dtype=int),
                         'exog_3': pd.Categorical(range(100, 110))})
    
    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=3)
    results = forecaster._create_train_X_y(y=y, exog=exog)        
    expected = (
        pd.DataFrame(
            data = np.array([[4, 3, 2, 1, 0, 105, 1005, 105, 106, 1006, 106, 107, 1007, 107],
                             [5, 4, 3, 2, 1, 106, 1006, 106, 107, 1007, 107, 108, 1008, 108],
                             [6, 5, 4, 3, 2, 107, 1007, 107, 108, 1008, 108, 109, 1009, 109]],
                            dtype=float),
            index   = pd.RangeIndex(start=7, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'exog_1_step_1', 'exog_2_step_1', 'exog_3_step_1', 
                       'exog_1_step_2', 'exog_2_step_2', 'exog_3_step_2', 
                       'exog_1_step_3', 'exog_2_step_3', 'exog_3_step_3']
        ).astype({'exog_1_step_1': float, 'exog_2_step_1': int,
                  'exog_1_step_2': float, 'exog_2_step_2': int,
                  'exog_1_step_3': float, 'exog_2_step_3': int}
        ).assign(exog_3_step_1=pd.Categorical(range(105, 108), categories=range(100, 110)),
                 exog_3_step_2=pd.Categorical(range(106, 109), categories=range(100, 110)),
                 exog_3_step_3=pd.Categorical(range(107, 110), categories=range(100, 110))
        ),
        {1: pd.Series(
                data  = np.array([5., 6., 7.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=8, step=1),
                name  = "y_step_1"
            ),
         2: pd.Series(
                data  = np.array([6., 7., 8.], dtype=float), 
                index = pd.RangeIndex(start=6, stop=9, step=1),
                name  = "y_step_2"
            ),
         3: pd.Series(
                data  = np.array([7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=7, stop=10, step=1),
                name  = "y_step_3"
            )
        },
        ['exog_1', 'exog_2', 'exog_3'],
        ['exog_1', 'exog_2', 'exog_3'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
         'exog_1_step_1', 'exog_2_step_1', 'exog_3_step_1', 
         'exog_1_step_2', 'exog_2_step_2', 'exog_3_step_2', 
         'exog_1_step_3', 'exog_2_step_3', 'exog_3_step_3'],
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes, 'exog_3': exog['exog_3'].dtypes},
        {'exog_1': exog['exog_1'].dtypes, 'exog_2': exog['exog_2'].dtypes, 'exog_3': exog['exog_3'].dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == [
        'exog_1_step_1', 'exog_2_step_1', 'exog_3_step_1', 
        'exog_1_step_2', 'exog_2_step_2', 'exog_3_step_2', 
        'exog_1_step_3', 'exog_2_step_3', 'exog_3_step_3'
    ]
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


def test_create_train_X_y_output_when_y_is_series_10_and_transformer_y_is_StandardScaler():
    """
    Test the output of _create_train_X_y when exog is None and transformer_y
    is StandardScaler with steps=1.
    """
    forecaster = ForecasterDirect(
                     regressor     = LinearRegression(),
                     lags          = 5,
                     steps         = 1,
                     transformer_y = StandardScaler()
                 )
    results = forecaster._create_train_X_y(y=pd.Series(np.arange(10), name='y', dtype=float))
    expected = (
        pd.DataFrame(
            data = np.array([[-0.17407766, -0.52223297, -0.87038828, -1.21854359, -1.5666989],
                             [0.17407766, -0.17407766, -0.52223297, -0.87038828, -1.21854359],
                             [0.52223297,  0.17407766, -0.17407766, -0.52223297, -0.87038828],
                             [0.87038828,  0.52223297,  0.17407766, -0.17407766, -0.52223297],
                             [1.21854359,  0.87038828,  0.52223297,  0.17407766, -0.17407766]]),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']
        ),
        {1: pd.Series(
                data  = np.array([0.17407766, 0.52223297, 0.87038828, 
                                  1.21854359, 1.5666989]), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "y_step_1"
            )
        },
        None,
        None,
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5'],
        None,
        None
    )

    forecaster.exog_in_ is False
    forecaster.X_train_direct_exog_names_out_ is None
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    assert results[5] == expected[5]
    assert results[6] == expected[6]


def test_create_train_X_y_output_when_lags_3_steps_1_and_exog_is_None_and_transformer_exog_is_not_None():
    """
    Test output of _create_train_X_y when regressor is LinearRegression, lags is 3
    and steps is 1 and transformer_exog is not None.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)

    forecaster = ForecasterDirect(
                     regressor        = LinearRegression(),
                     lags             = 3,
                     steps            = 1,
                     transformer_exog = StandardScaler()
                 )
    results = forecaster._create_train_X_y(y=y, exog=None)
    
    expected = (
        pd.DataFrame(
            data = np.array([[2., 1., 0.],
                             [3., 2., 1.],
                             [4., 3., 2.],
                             [5., 4., 3.],
                             [6., 5., 4.],
                             [7., 6., 5.],
                             [8., 7., 6.]], dtype=float),
            index   = pd.RangeIndex(start=3, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3']
        ),
        {1: pd.Series(
                data  = np.array([3., 4., 5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=3, stop=10, step=1),
                name  = "y_step_1"
            )
        },
        None,
        None,
        ['lag_1', 'lag_2', 'lag_3'],
        None,
        None
    )

    forecaster.exog_in_ is False
    forecaster.X_train_direct_exog_names_out_ is None
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    assert results[5] == expected[5]
    assert results[6] == expected[6]


def test_create_train_X_y_output_when_transformer_y_and_transformer_exog_steps_2():
    """
    Test the output of _create_train_X_y when using transformer_y and transformer_exog 
    with steps=2.
    """
    y = pd.Series(np.arange(10), dtype = float)
    y.index = pd.date_range("1990-01-01", periods=10, freq='D')
    exog = pd.DataFrame({
               'col_1': [7.5, 24.4, 54.2, 12.1, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4],
               'col_2': ['a', 'b', 'a', 'b', 'a', 'b', 'a', 'b', 'a', 'b']},
               index = pd.date_range("1990-01-01", periods=10, freq='D')
           )

    transformer_y = StandardScaler()
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['col_1']),
                            ('onehot', OneHotEncoder(), ['col_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )

    forecaster = ForecasterDirect(
                     regressor        = LinearRegression(),
                     lags             = 5,
                     steps            = 2,
                     transformer_y    = transformer_y,
                     transformer_exog = transformer_exog
                 )
    results = forecaster._create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[-0.17407766, -0.52223297, -0.87038828, -1.21854359, -1.5666989,
                              0.57176024, 0. , 1., 0.28259414, 1. , 0.],
                             [0.17407766, -0.17407766, -0.52223297, -0.87038828, -1.21854359,
                              0.28259414,  1. ,  0. , -0.12486718,  0. , 1. ],
                             [0.52223297,  0.17407766, -0.17407766, -0.52223297, -0.87038828,
                              -0.12486718,  0. ,  1. ,  1.88177028,  1. , 0. ],
                             [0.87038828,  0.52223297,  0.17407766, -0.17407766, -0.52223297,
                              1.88177028,  1. ,  0. ,  0.13801109,  0. , 1. ]]),
            index   = pd.date_range("1990-01-07", periods=4, freq='D'),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'col_1_step_1',
                       'col_2_a_step_1', 'col_2_b_step_1', 'col_1_step_2', 'col_2_a_step_2',
                       'col_2_b_step_2']
        ),
        {1: pd.Series(
                data  = np.array([0.17407766, 0.52223297, 0.87038828, 1.21854359]), 
                index = pd.date_range("1990-01-06", periods=4, freq='D'),
                name  = "y_step_1"
            ),
         2: pd.Series(
                data  = np.array([0.52223297, 0.87038828, 1.21854359, 1.5666989]), 
                index = pd.date_range("1990-01-07", periods=4, freq='D'),
                name  = "y_step_2"
            )
        },
        ['col_1', 'col_2'],
        ['col_1', 'col_2_a', 'col_2_b'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
         'col_1_step_1', 'col_2_a_step_1', 'col_2_b_step_1', 
         'col_1_step_2', 'col_2_a_step_2', 'col_2_b_step_2'],
        {'col_1': exog['col_1'].dtypes, 'col_2': exog['col_2'].dtypes},
        {'col_1': exog['col_1'].dtypes, 'col_2_a': float, 'col_2_b': float}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == [
        'col_1_step_1', 'col_2_a_step_1', 'col_2_b_step_1', 
        'col_1_step_2', 'col_2_a_step_2', 'col_2_b_step_2'
    ]
    forecaster.X_train_window_features_names_out_ is None

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


@pytest.mark.parametrize("fit_forecaster", 
                         [True, False], 
                         ids = lambda fitted: f'fit_forecaster: {fitted}')
def test_create_train_X_y_output_when_y_is_series_exog_is_series_and_differentiation_is_1(fit_forecaster):
    """
    Test the output of _create_train_X_y when using differentiation=1. Comparing 
    the matrix created with and without differentiating the series.
    """
    # Data differentiated
    differentiator = TimeSeriesDifferentiator(order=1)
    data_diff = differentiator.fit_transform(data.to_numpy())
    data_diff = pd.Series(data_diff, index=data.index).dropna()

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    exog_diff = exog.iloc[1:]
    end_train = '2003-03-01 23:59:00'

    forecaster_1 = ForecasterDirect(LinearRegression(), steps=1, lags=5)
    forecaster_2 = ForecasterDirect(LinearRegression(), steps=1, lags=5, differentiation=1)
    
    if fit_forecaster:
        forecaster_2.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])

    output_1 = forecaster_1._create_train_X_y(
                   y    = data_diff.loc[:end_train],
                   exog = exog_diff.loc[:end_train]
               )
    output_2 = forecaster_2._create_train_X_y(
                   y    = data.loc[:end_train],
                   exog = exog.loc[:end_train]
               )

    forecaster_1.exog_in_ is True
    forecaster_1.X_train_direct_exog_names_out_ == ['exog_step_1']
    forecaster_1.X_train_window_features_names_out_ is None

    forecaster_2.exog_in_ is True
    forecaster_2.X_train_direct_exog_names_out_ == ['exog_step_1']
    forecaster_2.X_train_window_features_names_out_ is None
    
    pd.testing.assert_frame_equal(output_1[0], output_2[0], check_names=True)
    for key in output_1[1]: 
        pd.testing.assert_series_equal(output_1[1][key], output_2[1][key], check_names=True) 
    assert output_1[2] == output_2[2]
    assert output_1[3] == output_2[3]
    assert output_1[4] == output_2[4]
    for k in output_1[5].keys():
        assert output_1[5][k] == output_2[5][k]
    for k in output_1[6].keys():
        assert output_1[6][k] == output_2[6][k]


@pytest.mark.parametrize("fit_forecaster", 
                         [True, False], 
                         ids = lambda fitted: f'fit_forecaster: {fitted}')
def test_create_train_X_y_output_when_y_is_series_exog_is_series_and_differentiation_is_1_steps_3(fit_forecaster):
    """
    Test the output of _create_train_X_y when using differentiation=1. Comparing 
    the matrix created with and without differentiating the series with steps=3.
    """
    # Data differentiated
    differentiator = TimeSeriesDifferentiator(order=1)
    data_diff = differentiator.fit_transform(data.to_numpy())
    data_diff = pd.Series(data_diff, index=data.index).dropna()

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    exog_diff = exog.iloc[1:]
    end_train = '2003-03-01 23:59:00'

    forecaster_1 = ForecasterDirect(LinearRegression(), steps=3, lags=5)
    forecaster_2 = ForecasterDirect(LinearRegression(), steps=3, lags=5, differentiation=1)
    
    if fit_forecaster:
        forecaster_2.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])

    output_1 = forecaster_1._create_train_X_y(
                   y    = data_diff.loc[:end_train],
                   exog = exog_diff.loc[:end_train]
               )
    output_2 = forecaster_2._create_train_X_y(
                   y    = data.loc[:end_train],
                   exog = exog.loc[:end_train]
               )

    forecaster_1.exog_in_ is True
    forecaster_1.X_train_direct_exog_names_out_ == ['exog_step_1', 'exog_step_2', 'exog_step_3']
    forecaster_1.X_train_window_features_names_out_ is None

    forecaster_2.exog_in_ is True
    forecaster_2.X_train_direct_exog_names_out_ == ['exog_step_1', 'exog_step_2', 'exog_step_3']
    forecaster_2.X_train_window_features_names_out_ is None
    
    pd.testing.assert_frame_equal(output_1[0], output_2[0], check_names=True)
    for key in output_1[1]: 
        pd.testing.assert_series_equal(output_1[1][key], output_2[1][key], check_names=True) 
    assert output_1[2] == output_2[2]
    assert output_1[3] == output_2[3]
    assert output_1[4] == output_2[4]
    for k in output_1[5].keys():
        assert output_1[5][k] == output_2[5][k]
    for k in output_1[6].keys():
        assert output_1[6][k] == output_2[6][k]


def test_create_train_X_y_output_when_y_is_series_exog_is_series_and_differentiation_is_2():
    """
    Test the output of _create_train_X_y when using differentiation=2. Comparing 
    the matrix created with and without differentiating the series.
    """

    # Data differentiated
    differentiator = TimeSeriesDifferentiator(order=2)
    data_diff_2 = differentiator.fit_transform(data.to_numpy())
    data_diff_2 = pd.Series(data_diff_2, index=data.index).dropna()

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    exog_diff_2 = exog.iloc[2:]
    end_train = '2003-03-01 23:59:00'

    forecaster_1 = ForecasterDirect(LinearRegression(), steps=1, lags=5)
    forecaster_2 = ForecasterDirect(
        LinearRegression(), steps=1, lags=5, differentiation=2
    )

    output_1 = forecaster_1._create_train_X_y(
                   y    = data_diff_2.loc[:end_train],
                   exog = exog_diff_2.loc[:end_train]
               )
    output_2 = forecaster_2._create_train_X_y(
                   y    = data.loc[:end_train],
                   exog = exog.loc[:end_train]
               )

    forecaster_1.exog_in_ is True
    forecaster_1.X_train_direct_exog_names_out_ == ['exog_step_1']
    forecaster_1.X_train_window_features_names_out_ is None

    forecaster_2.exog_in_ is True
    forecaster_2.X_train_direct_exog_names_out_ == ['exog_step_1']
    forecaster_2.X_train_window_features_names_out_ is None
    
    pd.testing.assert_frame_equal(output_1[0], output_2[0], check_names=True)
    for key in output_1[1]: 
        pd.testing.assert_series_equal(output_1[1][key], output_2[1][key], check_names=True) 
    assert output_1[2] == output_2[2]
    assert output_1[3] == output_2[3]
    assert output_1[4] == output_2[4]
    for k in output_1[5].keys():
        assert output_1[5][k] == output_2[5][k]
    for k in output_1[6].keys():
        assert output_1[6][k] == output_2[6][k]


def test_create_train_X_y_output_when_window_features_and_exog_steps_1():
    """
    Test the output of _create_train_X_y when using window_features and exog 
    with datetime index and steps=1.
    """
    y_datetime = pd.Series(
        np.arange(15), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='y', dtype=float
    )
    exog_datetime = pd.Series(
        np.arange(100, 115), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='exog', dtype=float
    )
    rolling = RollingFeatures(
        stats=['mean', 'median', 'sum'], window_sizes=[5, 5, 6]
    )

    forecaster = ForecasterDirect(
        LinearRegression(), steps=1, lags=5, window_features=rolling
    )
    results = forecaster._create_train_X_y(y=y_datetime, exog=exog_datetime)
    
    expected = (
        pd.DataFrame(
            data = np.array([[5., 4., 3., 2., 1., 3., 3., 15., 106.],
                             [6., 5., 4., 3., 2., 4., 4., 21., 107.],
                             [7., 6., 5., 4., 3., 5., 5., 27., 108.],
                             [8., 7., 6., 5., 4., 6., 6., 33., 109.],
                             [9., 8., 7., 6., 5., 7., 7., 39., 110.],
                             [10., 9., 8., 7., 6., 8., 8., 45., 111.],
                             [11., 10., 9., 8., 7., 9., 9., 51., 112.],
                             [12., 11., 10., 9., 8., 10., 10., 57., 113.],
                             [13., 12., 11., 10., 9., 11., 11., 63., 114.]]),
            index   = pd.date_range('2000-01-07', periods=9, freq='D'),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'roll_mean_5', 'roll_median_5', 'roll_sum_6', 'exog_step_1']
        ),
        {1: pd.Series(
                data  = np.array([6, 7, 8, 9, 10, 11, 12, 13, 14]),
                index = pd.date_range('2000-01-07', periods=9, freq='D'),
                name  = 'y_step_1',
                dtype = float
            )
        },
        ['exog'],
        ['exog'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
         'roll_mean_5', 'roll_median_5', 'roll_sum_6', 'exog_step_1'],
        {'exog': exog_datetime.dtypes},
        {'exog': exog_datetime.dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_step_1']
    forecaster.window_features_names == ['roll_mean_5', 'roll_median_5', 'roll_sum_6']
    forecaster.X_train_window_features_names_out_ == ['roll_mean_5', 'roll_median_5', 'roll_sum_6']

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


def test_create_train_X_y_output_when_window_features_and_exog_steps_2():
    """
    Test the output of _create_train_X_y when using window_features and exog 
    with datetime index and steps=2.
    """
    y_datetime = pd.Series(
        np.arange(15), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='y', dtype=float
    )
    exog_datetime = pd.Series(
        np.arange(100, 115), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='exog', dtype=float
    )
    rolling = RollingFeatures(
        stats=['mean', 'median', 'sum'], window_sizes=[5, 5, 6]
    )

    forecaster = ForecasterDirect(
        LinearRegression(), steps=2, lags=5, window_features=rolling
    )
    results = forecaster._create_train_X_y(y=y_datetime, exog=exog_datetime)
    
    expected = (
        pd.DataFrame(
            data = np.array([[5., 4., 3., 2., 1., 3., 3., 15., 106., 107.],
                             [6., 5., 4., 3., 2., 4., 4., 21., 107., 108.],
                             [7., 6., 5., 4., 3., 5., 5., 27., 108., 109.],
                             [8., 7., 6., 5., 4., 6., 6., 33., 109., 110.],
                             [9., 8., 7., 6., 5., 7., 7., 39., 110., 111.],
                             [10., 9., 8., 7., 6., 8., 8., 45., 111., 112.],
                             [11., 10., 9., 8., 7., 9., 9., 51., 112., 113.],
                             [12., 11., 10., 9., 8., 10., 10., 57., 113., 114.]]),
            index   = pd.date_range('2000-01-08', periods=8, freq='D'),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'roll_mean_5', 'roll_median_5', 'roll_sum_6', 
                       'exog_step_1', 'exog_step_2']
        ),
        {1: pd.Series(
                data  = np.array([6, 7, 8, 9, 10, 11, 12, 13]),
                index = pd.date_range('2000-01-07', periods=8, freq='D'),
                name  = 'y_step_1',
                dtype = float
            ),
         2: pd.Series(
                data  = np.array([7, 8, 9, 10, 11, 12, 13, 14]),
                index = pd.date_range('2000-01-08', periods=8, freq='D'),
                name  = 'y_step_2',
                dtype = float
            )   
        },
        ['exog'],
        ['exog'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
         'roll_mean_5', 'roll_median_5', 'roll_sum_6', 
         'exog_step_1', 'exog_step_2'],
        {'exog': exog_datetime.dtypes},
        {'exog': exog_datetime.dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_step_1', 'exog_step_2']
    forecaster.window_features_names == ['roll_mean_5', 'roll_median_5', 'roll_sum_6']
    forecaster.X_train_window_features_names_out_ == ['roll_mean_5', 'roll_median_5', 'roll_sum_6']

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


def test_create_train_X_y_output_when_two_window_features_and_exog_steps_2():
    """
    Test the output of _create_train_X_y when using 2 window_features and exog 
    with datetime index and steps=2.
    """
    y_datetime = pd.Series(
        np.arange(15), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='y', dtype=float
    )
    exog_datetime = pd.Series(
        np.arange(100, 115), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='exog', dtype=float
    )
    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=[5, 5])
    rolling_2 = RollingFeatures(stats='sum', window_sizes=[6])

    forecaster = ForecasterDirect(
        LinearRegression(), steps=2, lags=5, window_features=[rolling, rolling_2]
    )
    results = forecaster._create_train_X_y(y=y_datetime, exog=exog_datetime)
    
    expected = (
        pd.DataFrame(
            data = np.array([[5., 4., 3., 2., 1., 3., 3., 15., 106., 107.],
                             [6., 5., 4., 3., 2., 4., 4., 21., 107., 108.],
                             [7., 6., 5., 4., 3., 5., 5., 27., 108., 109.],
                             [8., 7., 6., 5., 4., 6., 6., 33., 109., 110.],
                             [9., 8., 7., 6., 5., 7., 7., 39., 110., 111.],
                             [10., 9., 8., 7., 6., 8., 8., 45., 111., 112.],
                             [11., 10., 9., 8., 7., 9., 9., 51., 112., 113.],
                             [12., 11., 10., 9., 8., 10., 10., 57., 113., 114.]]),
            index   = pd.date_range('2000-01-08', periods=8, freq='D'),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'roll_mean_5', 'roll_median_5', 'roll_sum_6', 
                       'exog_step_1', 'exog_step_2']
        ),
        {1: pd.Series(
                data  = np.array([6, 7, 8, 9, 10, 11, 12, 13]),
                index = pd.date_range('2000-01-07', periods=8, freq='D'),
                name  = 'y_step_1',
                dtype = float
            ),
         2: pd.Series(
                data  = np.array([7, 8, 9, 10, 11, 12, 13, 14]),
                index = pd.date_range('2000-01-08', periods=8, freq='D'),
                name  = 'y_step_2',
                dtype = float
            )   
        },
        ['exog'],
        ['exog'],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
         'roll_mean_5', 'roll_median_5', 'roll_sum_6', 
         'exog_step_1', 'exog_step_2'],
        {'exog': exog_datetime.dtypes},
        {'exog': exog_datetime.dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_step_1', 'exog_step_2']
    forecaster.window_features_names == ['roll_mean_5', 'roll_median_5', 'roll_sum_6']
    forecaster.X_train_window_features_names_out_ == ['roll_mean_5', 'roll_median_5', 'roll_sum_6']

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


def test_create_train_X_y_output_when_window_features_lags_None_and_exog():
    """
    Test the output of _create_train_X_y when using window_features and exog 
    with datetime index and lags=None and steps=1.
    """
    y_datetime = pd.Series(
        np.arange(15), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='y', dtype=float
    )
    exog_datetime = pd.Series(
        np.arange(100, 115), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='exog', dtype=float
    )
    rolling = RollingFeatures(
        stats=['mean', 'median', 'sum'], window_sizes=[5, 5, 6]
    )

    forecaster = ForecasterDirect(
        LinearRegression(), steps=1, lags=None, window_features=rolling
    )
    results = forecaster._create_train_X_y(y=y_datetime, exog=exog_datetime)
    
    expected = (
        pd.DataFrame(
            data = np.array([[3., 3., 15., 106.],
                             [4., 4., 21., 107.],
                             [5., 5., 27., 108.],
                             [6., 6., 33., 109.],
                             [7., 7., 39., 110.],
                             [8., 8., 45., 111.],
                             [9., 9., 51., 112.],
                             [10., 10., 57., 113.],
                             [11., 11., 63., 114.]]),
            index   = pd.date_range('2000-01-07', periods=9, freq='D'),
            columns = ['roll_mean_5', 'roll_median_5', 'roll_sum_6', 
                       'exog_step_1']
        ),
        {1: pd.Series(
                data  = np.array([6, 7, 8, 9, 10, 11, 12, 13, 14]),
                index = pd.date_range('2000-01-07', periods=9, freq='D'),
                name  = 'y_step_1',
                dtype = float
            )
        },
        ['exog'],
        ['exog'],
        ['roll_mean_5', 'roll_median_5', 'roll_sum_6', 'exog_step_1'],
        {'exog': exog_datetime.dtypes},
        {'exog': exog_datetime.dtypes}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['exog_step_1']
    forecaster.window_features_names == ['roll_mean_5', 'roll_median_5', 'roll_sum_6']
    forecaster.X_train_window_features_names_out_ == ['roll_mean_5', 'roll_median_5', 'roll_sum_6']

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


def test_create_train_X_y_output_when_window_features_and_exog_transformers_diff():
    """
    Test the output of _create_train_X_y when using window_features, exog, 
    transformers and differentiation with steps=1.
    """
    y_datetime = pd.Series(
        [25.3, 29.1, 27.5, 24.3, 2.1, 46.5, 31.3, 87.1, 133.5, 4.3],
        index=pd.date_range('2000-01-01', periods=10, freq='D'),
        name='y', dtype=float
    )
    exog = pd.DataFrame({
        'col_1': [7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4, 14.6, 73.5],
        'col_2': ['a', 'a', 'a', 'a', 'a', 'b', 'b', 'b', 'b', 'b']},
        index = pd.date_range('2000-01-01', periods=10, freq='D')
    )

    transformer_y = StandardScaler()
    transformer_exog = ColumnTransformer(
                            [('scale', StandardScaler(), ['col_1']),
                             ('onehot', OneHotEncoder(), ['col_2'])],
                            remainder = 'passthrough',
                            verbose_feature_names_out = False
                        )
    rolling = RollingFeatures(
        stats=['ratio_min_max', 'median'], window_sizes=4
    )

    forecaster = ForecasterDirect(
                     LinearRegression(), 
                     steps            = 1,
                     lags             = [1, 5], 
                     window_features  = rolling,
                     transformer_y    = transformer_y,
                     transformer_exog = transformer_exog,
                     differentiation  = 2
                 )
    results = forecaster._create_train_X_y(y=y_datetime, exog=exog)
    
    expected = (
        pd.DataFrame(
            data = np.array([[-1.56436158, -0.14173746, -0.89489489, -0.27035108,  0.04040264,
                               0.        ,  1.        ],
                             [ 1.8635851 , -0.04199628, -0.83943662,  0.62469472, -1.32578962,
                               0.        ,  1.        ],
                             [-0.24672817, -0.49870587, -0.83943662,  0.75068358,  1.12752513,
                               0.        ,  1.        ]]),
            index   = pd.date_range('2000-01-08', periods=3, freq='D'),
            columns = ['lag_1', 'lag_5', 'roll_ratio_min_max_4', 'roll_median_4',
                       'col_1_step_1', 'col_2_a_step_1', 'col_2_b_step_1']
        ),
        {1: pd.Series(
                data  = np.array([1.8635851, -0.24672817, -4.60909217]),
                index = pd.date_range('2000-01-08', periods=3, freq='D'),
                name  = 'y_step_1',
                dtype = float
            )
        },
        ['col_1', 'col_2'],
        ['col_1', 'col_2_a', 'col_2_b'],
        ['lag_1', 'lag_5', 'roll_ratio_min_max_4', 'roll_median_4',
         'col_1_step_1', 'col_2_a_step_1', 'col_2_b_step_1'],
        {'col_1': exog['col_1'].dtypes, 'col_2': exog['col_2'].dtypes},
        {'col_1': exog['col_1'].dtypes, 'col_2_a': float, 'col_2_b': float}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == ['col_1_step_1', 'col_2_a_step_1', 'col_2_b_step_1']
    forecaster.window_features_names == ['roll_ratio_min_max_4', 'roll_median_4']
    forecaster.X_train_window_features_names_out_ == ['roll_ratio_min_max_4', 'roll_median_4']

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]


def test_create_train_X_y_output_when_window_features_and_exog_transformers_diff_steps_2():
    """
    Test the output of _create_train_X_y when using window_features, exog, 
    transformers and differentiation with steps=2.
    """
    y_datetime = pd.Series(
        [25.3, 29.1, 27.5, 24.3, 2.1, 46.5, 31.3, 87.1, 133.5, 4.3],
        index=pd.date_range('2000-01-01', periods=10, freq='D'),
        name='y', dtype=float
    )
    exog = pd.DataFrame({
               'col_1': [7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4, 14.6, 73.5],
               'col_2': ['a', 'a', 'a', 'a', 'a', 'b', 'b', 'b', 'b', 'b']},
               index = pd.date_range('2000-01-01', periods=10, freq='D')
           )

    transformer_y = StandardScaler()
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['col_1']),
                            ('onehot', OneHotEncoder(), ['col_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )
    rolling = RollingFeatures(
        stats=['ratio_min_max', 'median'], window_sizes=4
    )

    forecaster = ForecasterDirect(
                     LinearRegression(), 
                     steps            = 2,
                     lags             = [1, 5], 
                     window_features  = rolling,
                     transformer_y    = transformer_y,
                     transformer_exog = transformer_exog,
                     differentiation  = 1
                 )
    results = forecaster._create_train_X_y(y=y_datetime, exog=exog)
    
    expected = (
        pd.DataFrame(
            data = np.array([
                       [ 1.16539688,  0.09974117, -0.5       , -0.06299443,  1.69816032,
                         0.        ,  1.        ,  0.04040264,  0.        ,  1.        ],
                       [-0.3989647 , -0.04199628, -0.5       , -0.24147863,  0.04040264,
                         0.        ,  1.        , -1.32578962,  0.        ,  1.        ],
                       [ 1.46462041, -0.08399257, -0.39784946,  0.38321609, -1.32578962,
                         0.        ,  1.        ,  1.12752513,  0.        ,  1.        ]]
                   ),
            index   = pd.date_range('2000-01-08', periods=3, freq='D'),
            columns = ['lag_1', 'lag_5', 'roll_ratio_min_max_4', 'roll_median_4',
                       'col_1_step_1', 'col_2_a_step_1', 'col_2_b_step_1', 
                       'col_1_step_2', 'col_2_a_step_2', 'col_2_b_step_2']
        ),
        {1: pd.Series(
                data  = np.array([-0.3989647, 1.46462041, 1.21789224]),
                index = pd.date_range('2000-01-07', periods=3, freq='D'),
                name  = 'y_step_1',
                dtype = float
            ),
         2: pd.Series(
                data  = np.array([1.46462041, 1.21789224, -3.39119993]),
                index = pd.date_range('2000-01-08', periods=3, freq='D'),
                name  = 'y_step_2',
                dtype = float
            )
        },
        ['col_1', 'col_2'],
        ['col_1', 'col_2_a', 'col_2_b'],
        ['lag_1', 'lag_5', 'roll_ratio_min_max_4', 'roll_median_4',
         'col_1_step_1', 'col_2_a_step_1', 'col_2_b_step_1', 
         'col_1_step_2', 'col_2_a_step_2', 'col_2_b_step_2'],
        {'col_1': exog['col_1'].dtypes, 'col_2': exog['col_2'].dtypes},
        {'col_1': exog['col_1'].dtypes, 'col_2_a': float, 'col_2_b': float}
    )

    forecaster.exog_in_ is True
    forecaster.X_train_direct_exog_names_out_ == [
        'col_1_step_1', 'col_2_a_step_1', 'col_2_b_step_1', 
        'col_1_step_2', 'col_2_a_step_2', 'col_2_b_step_2'
    ]
    forecaster.window_features_names == ['roll_ratio_min_max_4', 'roll_median_4']
    forecaster.X_train_window_features_names_out_ == ['roll_ratio_min_max_4', 'roll_median_4']

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 
    assert results[2] == expected[2]
    assert results[3] == expected[3]
    assert results[4] == expected[4]
    for k in results[5].keys():
        assert results[5][k] == expected[5][k]
    for k in results[6].keys():
        assert results[6][k] == expected[6][k]



# Unit test initialize_window_features
# ==============================================================================
import re
import pytest
from skforecast.utils import initialize_window_features


class WindowFeature:  # pragma: no cover
    def __init__(self, window_sizes, features_names):
        self.window_sizes = window_sizes
        self.features_names = features_names

    def transform_batch(self):
        pass

    def transform(self):
        pass


class InvalidWindowFeatureNoAttributes:  # pragma: no cover
    def transform_batch(self):
        pass

    def transform(self):
        pass


class InvalidWindowFeatureNoMethods:  # pragma: no cover
    def __init__(self, window_sizes, features_names):
        self.window_sizes = window_sizes
        self.features_names = features_names


link_to_docs = (
    "\nVisit the documentation for more information about how to create "
    "custom window features:\n"
    "https://skforecast.org/latest/user_guides/window-features-and-custom-features.html#create-your-custom-window-features"
)


def test_ValueError_initialize_window_features_when_empty_list():
    """
    Test ValueError is raised when `window_features` is an empty list.
    """
    err_msg = re.escape(
        "Argument `window_features` must contain at least one element."
    )
    with pytest.raises(ValueError, match = err_msg):
        initialize_window_features(window_features=[])


def test_ValueError_initialize_window_features_when_no_required_attributes():
    """
    Test ValueError is raised when window feature does not have the required attributes.
    """
    err_msg = re.escape(
        "InvalidWindowFeatureNoAttributes must have the attributes: "
        "['window_sizes', 'features_names']." + link_to_docs
    )
    with pytest.raises(ValueError, match = err_msg):
        initialize_window_features(InvalidWindowFeatureNoAttributes())


def test_ValueError_initialize_window_features_when_no_required_methods():
    """
    Test ValueError is raised when window feature does not have the required attributes.
    """
    err_msg = re.escape(
        "InvalidWindowFeatureNoMethods must have the methods: "
        "['transform_batch', 'transform']."  + link_to_docs
    )
    with pytest.raises(ValueError, match = err_msg):
        initialize_window_features(InvalidWindowFeatureNoMethods(5, 'feature_1'))


def test_TypeError_initialize_window_features_when_window_sizes_not_int_list():
    """
    Test TypeError is raised when `window_sizes` is not an int or a list of ints.
    """
    window_sizes = 'not_valid'

    err_msg = re.escape(
        f"Attribute `window_sizes` of WindowFeature must be an int or a list "
        f"of ints. Got {type(window_sizes)}."  + link_to_docs
    )
    with pytest.raises(TypeError, match = err_msg):
        initialize_window_features(WindowFeature(window_sizes, 'feature_1'))


def test_ValueError_initialize_window_features_when_window_sizes_int_lower_than_1():
    """
    Test ValueError is raised when `window_sizes` is an int lower than 1.
    """
    err_msg = re.escape(
        "If argument `window_sizes` is an integer, it must be equal "
        "to or greater than 1. Got 0 from WindowFeature."  + link_to_docs
    )
    with pytest.raises(ValueError, match = err_msg):
        initialize_window_features(WindowFeature(0, 'feature_1'))


@pytest.mark.parametrize("window_sizes", 
                         [[1, 2, 1.], [1, 3, 0], ['1', 3, 5]], 
                         ids = lambda window_sizes: f'window_sizes: {window_sizes}')
def test_ValueError_initialize_window_features_when_window_sizes_list_not_int_or_lower_than_1(window_sizes):
    """
    Test ValueError is raised when `window_sizes` is a list with elements not 
    int or lower than 1.
    """
    err_msg = re.escape(
        f"If argument `window_sizes` is a list, all elements must be integers "
        f"equal to or greater than 1. Got {window_sizes} from WindowFeature." + link_to_docs
    )
    with pytest.raises(ValueError, match = err_msg):
        initialize_window_features(WindowFeature(window_sizes, 'feature_1'))


def test_TypeError_initialize_window_features_when_features_names_not_str_list():
    """
    Test TypeError is raised when `features_names` is not a str or a list of strs.
    """
    features_names = 5

    err_msg = re.escape(
        f"Attribute `features_names` of WindowFeature must be a str or a list "
        f"of strings. Got {type(features_names)}."  + link_to_docs
    )
    with pytest.raises(TypeError, match = err_msg):
        initialize_window_features(WindowFeature(3, features_names))


def test_TypeError_initialize_window_features_when_features_names_list_not_str():
    """
    Test TypeError is raised when `features_names` is a list with elements not str.
    """
    features_names = ['name_1', 5]

    err_msg = re.escape(
        "If argument `features_names` is a list, all elements "
        "must be strings. Got ['name_1', 5] from WindowFeature." + link_to_docs
    )
    with pytest.raises(TypeError, match = err_msg):
        initialize_window_features(WindowFeature(3, features_names))


def test_ValueError_initialize_window_features_when_not_unique_features_names():
    """
    Test ValueError is raised when `features_names` are not unique.
    """
    window_features = [
        WindowFeature([4, 6], ['feature_1', 'feature_2']),
        WindowFeature(5, ['feature_2', 'feature_3'])
    ]
    window_features_names = ['feature_1', 'feature_2', 'feature_2', 'feature_3']

    err_msg = re.escape(
        f"All window features names must be unique. Got {window_features_names}."
    )
    with pytest.raises(ValueError, match = err_msg):
        initialize_window_features(window_features)


def test_initialize_window_features_when_None():
    """
    Test initialize_window_features with None.
    """

    window_features, window_features_names, max_size_window_features = (
        initialize_window_features(None)
    )

    assert window_features is None
    assert window_features_names is None
    assert max_size_window_features is None


def test_initialize_window_features_valid():
    """
    Test initialize_window_features with valid values.
    """

    wf1 = WindowFeature(5, "feature1")
    wf2 = WindowFeature([3, 4], ["feature2", "feature3"])
    window_features, window_features_names, max_size_window_features = (
        initialize_window_features([wf1, wf2])
    )

    assert window_features == [wf1, wf2]
    assert window_features_names == ["feature1", "feature2", "feature3"]
    assert max_size_window_features == 5



# Unit test fit method with TensorFlow backend
# ==============================================================================
import sys
import pytest

pytestmark = pytest.mark.skipif(
    sys.version_info >= (3, 13, 0),
    reason="TensorFlow does not support Python 3.13+",
)

import os
import numpy as np
import pandas as pd
os.environ["KERAS_BACKEND"] = "tensorflow"
import keras
from keras.optimizers import Adam
from keras.losses import MeanSquaredError
from skforecast.deep_learning import ForecasterRnn
from skforecast.deep_learning.utils import create_and_compile_model

series = pd.DataFrame(
    {"1": pd.Series(np.arange(5)), "2": pd.Series(np.arange(5))}
)
exog = pd.DataFrame(
    {"exog1": pd.Series(np.arange(5)), "exog2": pd.Series(np.arange(5))}
)
lags = 3
steps = 1
levels = "1"

model = create_and_compile_model(
    series=series, 
    levels=levels,    
    lags=lags,           
    steps=steps,              
    recurrent_layer="LSTM",
    recurrent_units=100,
    recurrent_layers_kwargs={"activation": "relu"},
    dense_units=[128, 64],
    dense_layers_kwargs={"activation": "relu"},
    output_dense_layer_kwargs={"activation": "linear"},
    compile_kwargs={"optimizer": Adam(learning_rate=0.01), "loss": MeanSquaredError()},
)

model_exog = create_and_compile_model(
    series=series, 
    exog=exog,
    levels=levels,    
    lags=lags,           
    steps=steps,              
    recurrent_layer="LSTM",
    recurrent_units=128,
    dense_units=64,
)


# Test case for fitting the forecaster without validation data
def test_fit_without_validation_data():
    """
    Test case for fitting the forecaster without validation data
    """
    # Call the function to create and compile the model

    forecaster = ForecasterRnn(model, levels, lags=lags)

    # Assert that the forecaster is fitted
    assert forecaster.is_fitted is False

    # Fit the forecaster
    forecaster.fit(series)

    # print(f"GITHUB_ACTIONS: {os.getenv('GITHUB_ACTIONS')}")
    # print(f"CI: {os.getenv('CI')}")
    # print(f"GITHUB_WORKFLOW in environ: {'GITHUB_WORKFLOW' in os.environ}")
    # print(f"GITHUB_WORKFLOW value: {os.getenv('GITHUB_WORKFLOW')}")

    # Assert that the forecaster is fitted
    assert forecaster.is_fitted is True
    if not os.getenv('GITHUB_ACTIONS') == 'true':
        assert forecaster.keras_backend_ == "tensorflow"

    assert forecaster.series_names_in_ == ["1", "2"]
    assert forecaster.X_train_series_names_in_ == ["1", "2"]

    # Assert that the training range is set correctly
    assert all(forecaster.training_range_ == (0, 4))

    # Assert that the last window is set correctly
    last_window = pd.DataFrame({"1": [2, 3, 4], "2": [2, 3, 4]})

    np.testing.assert_array_almost_equal(forecaster.last_window_, last_window)


# Test case for fitting the forecaster with validation data
def test_fit_with_validation_data():
    """
    Test case for fitting the forecaster with validation data
    """

    # Create a validation series
    series_val = pd.DataFrame(
        {"1": pd.Series(np.arange(5)), "2": pd.Series(np.arange(5))}
    )

    # Create an instance of ForecasterRnn
    forecaster = ForecasterRnn(
        regressor=model,
        levels=levels,
        fit_kwargs={
            "epochs": 10,  # Number of epochs to train the model.
            "batch_size": 32,  # Batch size to train the model.
            "series_val": series_val,  # Validation data for model training.
        },
        lags=lags
    )

    # Assert that the forecaster is not fitted
    assert forecaster.is_fitted is False

    # Fit the forecaster
    forecaster.fit(series)

    # Assert that the forecaster is fitted
    assert forecaster.is_fitted is True
    if not os.getenv('GITHUB_ACTIONS') == 'true':
        assert forecaster.keras_backend_ == "tensorflow"

    # Assert that the history is not None
    assert forecaster.history_ is not None


def test_fit_with_exog_and_validation_data():
    """
    Test case for fitting the forecaster with validation data
    """

    series_val = pd.DataFrame(
        {"1": pd.Series(np.arange(5)), "2": pd.Series(np.arange(5))}
    )
    exog_val = pd.DataFrame(
        {"exog1": pd.Series(np.arange(5)), "exog2": pd.Series(np.arange(5))}
    )
    forecaster = ForecasterRnn(
        regressor=model_exog,
        levels=levels,
        fit_kwargs={
            "epochs": 10,  # Number of epochs to train the model.
            "batch_size": 32,  # Batch size to train the model.
            "series_val": series_val,  # Validation data for model training.
            "exog_val": exog_val,  # Validation exogenous data for model training.
        },
        lags=lags
    )
    assert forecaster.is_fitted is False
    
    forecaster.fit(series, exog=exog)

    assert forecaster.is_fitted is True
    if not os.getenv('GITHUB_ACTIONS') == 'true':
        assert forecaster.keras_backend_ == "tensorflow"
    
    assert forecaster.history_ is not None
    
    assert forecaster.exog_names_in_ == ["exog1", "exog2"]



from .exceptions import (
    DataTypeWarning,
    DataTransformationWarning,
    IgnoredArgumentWarning,
    IndexWarning,
    InputTypeWarning,
    LongTrainingWarning,
    MissingExogWarning,
    MissingValuesWarning,
    OneStepAheadValidationWarning,
    ResidualsUsageWarning,
    UnknownLevelWarning,
    SaveLoadSkforecastWarning,
    SkforecastVersionWarning,
    warn_skforecast_categories,
    set_warnings_style
)


# Unit test align_series_and_exog_multiseries
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.exceptions import MissingValuesWarning
from skforecast.utils import (
    check_preprocess_series,
    check_preprocess_exog_multiseries,
    align_series_and_exog_multiseries
)
from skforecast.preprocessing import reshape_series_wide_to_long
from skforecast.recursive.tests.tests_forecaster_recursive_multiseries.fixtures_forecaster_recursive_multiseries import (
    series_wide_range,
    series_wide_dt,
    series_long_dt,
    series_dict_range,
    series_dict_dt,
    exog_wide_range,
    exog_wide_dt,
    exog_long_dt,
    exog_dict_range,
    exog_dict_dt
)


def test_output_align_series_and_exog_multiseries_when_series_is_DataFrame_and_exog_None():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    and `exog` is None.
    """
    series_dict, series_indexes = check_preprocess_series(series=series_dict_range)

    exog_dict = {'l1': None, 'l2': None}

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict = series_dict,
                                 exog_dict   = exog_dict
                             )

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_dict_range['l1'].to_numpy(),
                  index = pd.RangeIndex(start=0, stop=50),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_dict_range['l2'].to_numpy(),
                  index = pd.RangeIndex(start=0, stop=50),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {'l1': None, 'l2': None}
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
        assert isinstance(exog_dict[k], type(None))
    for k in exog_dict:
        assert exog_dict[k] == expected_exog_dict[k]


def test_output_align_series_and_exog_multiseries_when_series_is_DataFrame_different_lengths_and_exog_None():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    with series of different lengths and `exog` is None.
    """
    series_diff = series_wide_dt.copy()
    n_nans = 10
    series_diff.iloc[:n_nans, 0] = np.nan
    series_diff = reshape_series_wide_to_long(data=series_diff)

    series_dict, series_indexes = check_preprocess_series(series=series_diff)

    exog_dict = {'1': None, '2': None}

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict = series_dict,
                                 exog_dict   = exog_dict
                             )

    expected_series_dict = {
        '1': pd.Series(
                 data  = series_wide_dt['1'].to_numpy()[-40:],
                 index = pd.date_range(start='2000-01-11', periods=len(series_wide_dt) - n_nans, freq='D'),
                 name  = '1'
             ),
        '2': pd.Series(
                 data  = series_wide_dt['2'].to_numpy(),
                 index = pd.date_range(start='2000-01-01', periods=len(series_wide_dt), freq='D'),
                 name  = '2'
             ),
    }
    expected_exog_dict = {'1': None, '2': None}
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['1', '2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
        assert isinstance(exog_dict[k], type(None))
    for k in exog_dict:
        assert exog_dict[k] == expected_exog_dict[k]


def test_output_align_series_and_exog_multiseries_when_series_is_DataFrame_different_lengths_and_nans_in_between():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    with series of different lengths and NaNs in between with `exog` is None.
    """
    series_diff = series_wide_dt.copy()
    n_nans = 10
    series_diff.iloc[:n_nans, 0] = np.nan
    series_diff.iloc[20:23, 0] = np.nan
    series_diff.iloc[33:49, 1] = np.nan
    series_diff_long = reshape_series_wide_to_long(data=series_diff)

    series_dict, series_indexes = check_preprocess_series(series=series_diff_long)

    exog_dict = {'1': None, '2': None}

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict = series_dict,
                                 exog_dict   = exog_dict
                             )

    expected_series_dict = {
        '1': pd.Series(
                 data  = series_diff['1'].to_numpy()[-40:],
                 index = pd.date_range(start='2000-01-11', periods=len(series_wide_dt) - n_nans, freq='D'),
                 name  = '1'
             ),
        '2': pd.Series(
                 data  = series_diff['2'].to_numpy(),
                 index = pd.date_range(start='2000-01-01', periods=len(series_wide_dt), freq='D'),
                 name  = '2'
             ),
    }
    expected_exog_dict = {'1': None, '2': None}
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['1', '2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
        assert isinstance(exog_dict[k], type(None))
    for k in exog_dict:
        assert exog_dict[k] == expected_exog_dict[k]


def test_output_align_series_and_exog_multiseries_when_series_is_dict_different_lengths_and_nans_in_between():
    """
    Test align_series_and_exog_multiseries when `series` is a dict 
    with series of different lengths and NaNs in between with `exog` is None.
    """
    series_diff = {
        'l1': series_dict_dt['l1'].copy(),
        'l2': series_dict_dt['l2'].copy()
    }
    n_nans = 10
    series_diff['l1'].iloc[:n_nans, ] = np.nan
    series_diff['l1'].iloc[20:23, ] = np.nan
    series_diff['l2'].iloc[33:49, ] = np.nan

    series_dict, series_indexes = check_preprocess_series(series=series_diff)

    exog_dict = {'l1': None, 'l2': None}

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict = series_dict,
                                 exog_dict   = exog_dict
                             )

    len_series = len(series_dict_dt['l1'])
    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_diff['l1'].to_numpy()[-40:],
                  index = pd.date_range(start='2000-01-11', periods=len_series - n_nans, freq='D'),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_diff['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len_series, freq='D'),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {'l1': None, 'l2': None}
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
        assert isinstance(exog_dict[k], type(None))
    for k in exog_dict:
        assert exog_dict[k] == expected_exog_dict[k]


def test_output_align_series_and_exog_multiseries_when_input_series_is_DataFrame_and_exog_DataFrame():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    and `exog` is a pandas DataFrame.
    """

    series_dict, series_indexes = check_preprocess_series(series=series_long_dt)

    exog_long = exog_long_dt.copy()
    exog_long.index = exog_long.index.set_levels(['1', '2'], level='series_id')

    exog_dict, _ = check_preprocess_exog_multiseries(
                       series_names_in_  = ['1', '2'],
                       series_index_type = type(series_indexes['1']),
                       exog              = exog_long,
                       exog_dict         = {'1': None, '2': None}
                   )

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict = series_dict,
                                 exog_dict   = exog_dict
                             )

    len_series = len(series_dict_dt['l1'])
    expected_series_dict = {
        '1': pd.Series(
                 data  = series_dict_dt['l1'].to_numpy(),
                 index = pd.date_range(start='2000-01-01', periods=len_series, freq='D'),
                 name  = '1'
             ),
        '2': pd.Series(
                 data  = series_dict_dt['l2'].to_numpy(),
                 index = pd.date_range(start='2000-01-01', periods=len_series, freq='D'),
                 name  = '2'
             ),
    }
    expected_exog_dict = {
        '1': pd.DataFrame(
                 data    = exog_wide_range.to_numpy(),
                 index   = pd.date_range(start='2000-01-01', periods=len_series, freq='D'),
                 columns = ['exog_1', 'exog_2']
             ).astype({'exog_1': float}),
        '2': pd.DataFrame(
                 data    = exog_wide_range.to_numpy(),
                 index   = pd.date_range(start='2000-01-01', periods=len_series, freq='D'),
                 columns = ['exog_1', 'exog_2']
             ).astype({'exog_1': float}),
    }
    expected_exog_dict['1'].index.freq = None
    expected_exog_dict['2'].index.freq = None
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['1', '2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['1', '2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_input_series_is_DataFrame_and_exog_dict():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    and `exog` is a dict.
    """
    series_dict, series_indexes = check_preprocess_series(series=series_dict_range)

    exog_dict, _ = check_preprocess_exog_multiseries(
                       series_names_in_  = ['l1', 'l2'],
                       series_index_type = type(series_indexes['l1']),
                       exog              = exog_dict_range,
                       exog_dict         = {'l1': None, 'l2': None}
                   )

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict = series_dict,
                                 exog_dict   = exog_dict
                             )

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_dict_range['l1'].to_numpy(),
                  index = pd.RangeIndex(start=0, stop=50),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_dict_range['l2'].to_numpy(),
                  index = pd.RangeIndex(start=0, stop=50),
                  name  = 'l2'
             ),
    }
    expected_exog_dict = {
        'l1': pd.DataFrame(
                  data    = exog_wide_range.to_numpy(),
                  index   = pd.RangeIndex(start=0, stop=50),
                  columns = ['exog_1', 'exog_2']
              ).astype({'exog_1': float}),
        'l2': pd.DataFrame(
                  data    = exog_wide_range['exog_1'].to_numpy(),
                  index   = pd.RangeIndex(start=0, stop=50),
                  columns = ['exog_1']
              ).astype({'exog_1': float}),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['l1', 'l2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_series_is_DataFrame_different_lengths_and_nans_in_between_and_exog_dict():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    with series of different lengths and NaNs in between with `exog` is a 
    dict with no datetime index.
    """
    series_diff = series_wide_dt.copy()
    n_nans = 10
    series_diff.iloc[:n_nans, 0] = np.nan
    series_diff.iloc[20:23, 0] = np.nan
    series_diff.iloc[33:49, 1] = np.nan
    series_diff_long = reshape_series_wide_to_long(data=series_diff)

    exog_dict = {
        '1': exog_wide_dt.copy(),
        '2': exog_wide_dt['exog_1'].copy()
    }

    series_dict, series_indexes = check_preprocess_series(series=series_diff_long)

    exog_dict, _ = check_preprocess_exog_multiseries(
                       series_names_in_  = ['1', '2'],
                       series_index_type = type(series_indexes['1']),
                       exog              = exog_dict,
                       exog_dict         = {'1': None, '2': None}
                   )

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict = series_dict,
                                 exog_dict   = exog_dict
                             )

    expected_series_dict = {
        '1': pd.Series(
                 data  = series_diff['1'].to_numpy()[-40:],
                 index = pd.date_range(start='2000-01-11', periods=len(series_wide_dt) - n_nans, freq='D'),
                 name  = '1'
             ),
        '2': pd.Series(
                 data  = series_diff['2'].to_numpy(),
                 index = pd.date_range(start='2000-01-01', periods=len(series_wide_dt), freq='D'),
                 name  = '2'
             ),
    }
    expected_exog_dict = {
        '1': pd.DataFrame(
                 data    = exog_wide_range.to_numpy()[-40:],
                 index   = pd.date_range(start='2000-01-11', periods=len(series_wide_dt) - n_nans, freq='D'),
                 columns = ['exog_1', 'exog_2']
             ).astype({'exog_1': float}),
        '2': pd.DataFrame(
                 data    = exog_wide_range['exog_1'].to_numpy(),
                 index   = pd.date_range(start='2000-01-01', periods=len(series_wide_dt), freq='D'),
                 columns = ['exog_1']
             ).astype({'exog_1': float}),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['1', '2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['1', '2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_series_is_DataFrame_different_lengths_and_nans_in_between_and_exog_dict_datetime():
    """
    Test align_series_and_exog_multiseries when `series` is a pandas DataFrame 
    with series of different lengths and NaNs in between with `exog` is a 
    dict with datetime index.
    """
    series_diff = series_wide_dt.copy()
    n_nans = 10
    series_diff.iloc[:n_nans, 0] = np.nan
    series_diff.iloc[20:23, 0] = np.nan
    series_diff.iloc[33:49, 1] = np.nan
    series_diff.columns = ['l1', 'l2']
    series_diff_long = reshape_series_wide_to_long(data=series_diff)

    series_dict, series_indexes = check_preprocess_series(series=series_diff_long)

    exog_dict, _ = check_preprocess_exog_multiseries(
                       series_index_type = type(series_indexes['l1']),
                       series_names_in_  = ['l1', 'l2'],
                       exog              = exog_dict_dt,
                       exog_dict         = {'l1': None, 'l2': None}
                   )

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict = series_dict,
                                 exog_dict   = exog_dict
                             )

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_diff['l1'].to_numpy()[-40:],
                  index = pd.date_range(start='2000-01-11', periods=len(series_wide_dt) - n_nans, freq='D'),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_diff['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_wide_dt), freq='D'),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {
        'l1': pd.DataFrame(
                  data    = exog_dict_dt['l1'].to_numpy()[-40:],
                  index   = pd.date_range(start='2000-01-11', periods=len(series_wide_dt) - n_nans, freq='D'),
                  columns = ['exog_1', 'exog_2']
              ).astype({'exog_1': float}),
        'l2': pd.DataFrame(
                  data    = exog_dict_dt['l2'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(series_wide_dt), freq='D'),
                  columns = ['exog_1']
              ).astype({'exog_1': float}),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['l1', 'l2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_series_dict_and_exog_dict_datetime():
    """
    Test align_series_and_exog_multiseries when `series` is a dict and `exog` is 
    a dict. Datetime index is a must.
    """
    series_dict, series_indexes = check_preprocess_series(series=series_dict_dt)

    exog_dict, _ = check_preprocess_exog_multiseries(
                       series_names_in_  = ['l1', 'l2'],
                       series_index_type = type(series_indexes['l1']),
                       exog              = exog_dict_dt,
                       exog_dict         = {'l1': None, 'l2': None}
                   )

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict = series_dict,
                                 exog_dict   = exog_dict
                             )

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_dict_dt['l1'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_dict_dt['l1']), freq='D'),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_dict_dt['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_dict_dt['l2']), freq='D'),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {
        'l1': pd.DataFrame(
                  data    = exog_dict_dt['l1'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(exog_dict_dt['l1']), freq='D'),
                  columns = ['exog_1', 'exog_2']
              ).astype({'exog_1': float}),
        'l2': pd.DataFrame(
                  data    = exog_dict_dt['l2'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(exog_dict_dt['l2']), freq='D'),
                  columns = ['exog_1']
              ).astype({'exog_1': float}),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['l1', 'l2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_series_is_dict_different_lengths_and_nans_in_between_and_exog_dict_datetime():
    """
    Test align_series_and_exog_multiseries when `series` is a dict 
    with series of different lengths and NaNs in between with `exog` is a 
    dict with datetime index.
    """
    series_diff = {
        'l1': series_dict_dt['l1'].copy(),
        'l2': series_dict_dt['l2'].copy()
    }
    n_nans = 10
    series_diff['l1'].iloc[:n_nans, ] = np.nan
    series_diff['l1'].iloc[20:23, ] = np.nan
    series_diff['l2'].iloc[33:49, ] = np.nan

    series_dict, series_indexes = check_preprocess_series(series=series_diff)

    exog_dict, _ = check_preprocess_exog_multiseries(
                       series_names_in_  = ['l1', 'l2'],
                       series_index_type = type(series_indexes['l1']),
                       exog              = exog_dict_dt,
                       exog_dict         = {'l1': None, 'l2': None}
                   )

    series_dict, exog_dict = align_series_and_exog_multiseries(
                                 series_dict = series_dict,
                                 exog_dict   = exog_dict
                             )
    
    len_series = len(series_dict_dt['l1'])
    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_diff['l1'].to_numpy()[-40:],
                  index = pd.date_range(start='2000-01-11', periods=len_series - n_nans, freq='D'),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_diff['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len_series, freq='D'),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {
        'l1': pd.DataFrame(
                  data    = exog_dict_dt['l1'].to_numpy()[-40:],
                  index   = pd.date_range(start='2000-01-11', periods=len_series - n_nans, freq='D'),
                  columns = ['exog_1', 'exog_2']
              ).astype({'exog_1': float}),
        'l2': pd.DataFrame(
                  data    = exog_dict_dt['l2'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len_series, freq='D'),
                  columns = ['exog_1']
              ).astype({'exog_1': float}),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['l1', 'l2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_series_is_dict_and_length_intersection_with_exog_is_0():
    """
    Test align_series_and_exog_multiseries when `series` is a dict and the
    intersection with `exog` is 0.
    """
    series_diff = {
        'l1': series_dict_dt['l1'].copy(),
        'l2': series_dict_dt['l2'].copy()
    }
    n_nans = 10
    series_diff['l1'].iloc[:n_nans, ] = np.nan
    series_diff['l1'].iloc[20:23, ] = np.nan
    series_diff['l2'].iloc[33:49, ] = np.nan

    series_dict, series_indexes = check_preprocess_series(series=series_diff)

    exog_no_intersection = {
        'l1': exog_dict_dt['l1'].copy(),
        'l2': exog_dict_dt['l2'].copy()
    }
    exog_no_intersection['l1'].index = pd.date_range(
        start='2001-01-01', periods=len(exog_no_intersection['l1']), freq='D'
    )

    exog_dict, _ = check_preprocess_exog_multiseries(
                       series_names_in_  = ['l1', 'l2'],
                       series_index_type = type(series_indexes['l1']),
                       exog              = exog_no_intersection,
                       exog_dict         = {'l1': None, 'l2': None}
                   )

    warn_msg = re.escape(
        "`exog` for series 'l1' is empty after aligning "
        "with the series index. Exog values will be NaN."
    )
    with pytest.warns(MissingValuesWarning, match = warn_msg):
        series_dict, exog_dict = align_series_and_exog_multiseries(
                                     series_dict = series_dict,
                                     exog_dict   = exog_dict
                                 )

    len_series = len(series_dict_dt['l1'])
    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_diff['l1'].to_numpy()[-40:],
                  index = pd.date_range(start='2000-01-11', periods=len_series - n_nans, freq='D'),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_diff['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len_series, freq='D'),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {
        'l1': None,
        'l2': pd.DataFrame(
                  data    = exog_dict_dt['l2'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len_series, freq='D'),
                  columns = ['exog_1']
              ),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['l1', 'l2']
    for k in exog_dict:
        if exog_dict[k] is None:
            assert exog_dict[k] == expected_exog_dict[k]
        else:
            pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
            pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)


def test_output_align_series_and_exog_multiseries_when_series_is_dict_and_different_length_intersection_with_exog():
    """
    Test align_series_and_exog_multiseries when `series` is a dict and the
    intersection with `exog` do not have the same length as series.
    """
    series_diff = {
        'l1': series_dict_dt['l1'].copy(),
        'l2': series_dict_dt['l2'].copy()
    }
    n_nans = 10
    series_diff['l1'].iloc[:n_nans, ] = np.nan
    series_diff['l1'].iloc[20:23, ] = np.nan
    series_diff['l2'].iloc[33:49, ] = np.nan

    series_dict, series_indexes = check_preprocess_series(series=series_diff)

    exog_half_intersection = {
        'l1': exog_dict_dt['l1'].copy(),
        'l2': exog_dict_dt['l2'].copy()
    }
    exog_half_intersection['l1'].iloc[20:23, :] = np.nan
    exog_half_intersection['l1'].index = pd.date_range(
        start='2000-01-15', periods=len(exog_half_intersection['l1']), freq='D'
    )
    exog_half_intersection['l2'].iloc[:10] = np.nan
    exog_half_intersection['l2'].iloc[33:49] = np.nan

    exog_dict, _ = check_preprocess_exog_multiseries(
                       series_names_in_  = ['l1', 'l2'],
                       series_index_type = type(series_indexes['l1']),
                       exog              = exog_half_intersection,
                       exog_dict         = {'l1': None, 'l2': None}
                   )

    warn_msg = re.escape(
        "`exog` for series 'l1' doesn't have values for "
        "all the dates in the series. Missing values will be "
        "filled with NaN."
    )
    with pytest.warns(MissingValuesWarning, match = warn_msg):
        series_dict, exog_dict = align_series_and_exog_multiseries(
                                     series_dict = series_dict,
                                     exog_dict   = exog_dict
                                 )

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_diff['l1'].to_numpy()[-40:],
                  index = pd.date_range(start='2000-01-11', periods=len(series_dict_dt['l1']) - n_nans, freq='D'),
                  name  = 'l1'
              ),
        'l2': pd.Series(
                  data  = series_diff['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_dict_dt['l2']), freq='D'),
                  name  = 'l2'
              ),
    }
    expected_exog_dict = {
        'l1': pd.DataFrame(
                  data    = np.vstack((
                                np.full((4, len(['exog_1', 'exog_2'])), np.nan), 
                                exog_half_intersection['l1'].to_numpy()[:40 - 4]
                            )),
                  index   = pd.date_range(start='2000-01-11', periods=len(expected_series_dict['l1']), freq='D'),
                  columns = ['exog_1', 'exog_2']
              ).astype({'exog_1': float, 'exog_2': object}),
        'l2': pd.DataFrame(
                  data    = exog_half_intersection['l2'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(exog_half_intersection['l2']), freq='D'),
                  columns = ['exog_1']
              ).astype({'exog_1': float}),
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['l1', 'l2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)
